#ifndef __VECTOR_H__
#define __VECTOR_H__

#include <assert.h>
#include <math.h>

//CVectorNX
class CMatrix;
class CVector {
public:
	union {
		double	m_rElement[4];
		struct {
			union {
				double	m_rX;
				double	x;
			};
			union {
				double	m_rY;
				double	y;
			};
			union {
				double	m_rZ;
				double	z;
			};
			union {
				double	m_rW;
				double	w;
			};
		};
	};

public:
	CVector(){
		x=0.f;y=0.f;z=0.f;w=1.0f;
	}
	CVector(double X,double Y,double Z){//VisualC++5.0SP3𓖂ĂRXgN^ɃftHg̈tƓG[N悤ɐĂ܂߂̋̍;(
		x=X;y=Y;z=Z;w=1.0f;
	}
	CVector(double X,double Y,double Z,double W){
		x=X;y=Y;z=Z;w=W;
	}
	~CVector(){}
	double GetElement(int row) const{assert(0<=row && row<4);return m_rElement[row];}
	double GetX() const {return x;}
	double GetY() const {return y;}
	double GetZ() const {return z;}
	double GetW() const {return w;}
	void SetElement(int row, double val){assert(0<=row && row<4);m_rElement[row] = val;}
	void SetX(double val) {x = val;}
	void SetY(double val) {y = val;}
	void SetZ(double val) {z = val;}
	void SetW(double val) {w = val;}
	inline void Set(double X,double Y,double Z,double W=double(1));
	inline double GetLength();
		//_Ƃ̋
	inline CVector& Normalize();
		//PʃxNg
	inline CVector& Homogenize();
		//
	inline CVector& operator=(CVector const& vctr);
	inline CVector operator+(CVector const& vctr) const;
	inline CVector& operator+=(CVector const& vctr);
	inline CVector operator-(CVector const& vctr) const;
	inline CVector& operator-=(CVector const& vctr);
	inline CVector operator*(double val) const;
	inline CVector& operator*=(double val);
	inline CVector operator/(double val) const;
	inline CVector& operator/=(double val);
	inline BOOL operator==(CVector const& vctr) const;
	inline BOOL operator!=(CVector const& vctr) const;
	inline double DotProduct(CVector const& vctr) const;
		//
	inline CVector CrossProduct(CVector const& vctr) const;
		//O
	inline CVector operator*(CMatrix const& mtrx) const;
	inline CVector& operator*=(CMatrix const& mtrx);
	friend CVector operator*(double val, CVector const& vctr);
};
inline void CVector::Set(double X,double Y,double Z,double W/*=double(1)*/){
	x=X;
	y=Y;
	z=Z;
	w=W;
}

inline double CVector::GetLength(){
	return (double)sqrt(x*x +y*y +z*z);
}

inline CVector& CVector::Normalize(){
	double	len = GetLength();
	if(len!=double(0.0)){
		x /= len;
		y /= len;
		z /= len;
	}
	return *this;
}
inline CVector& CVector::Homogenize(){
	if(w!=double(0.0)){
		x /= m_rW;
		y /= m_rW;
		z /= m_rW;
		w = double(1.0);// m_rW /= m_W;
	}
	return *this;
}

inline CVector& CVector::operator=(CVector const& vctr)
{
	for(int i=0; i<4; i++){
		m_rElement[i] = vctr.GetElement(i);
	}
	return *this;
}

inline CVector CVector::operator+(CVector const& vctr) const
{
	// iKłwlĂȂ
	return CVector(
		x + vctr.x,
		y + vctr.y,
		z + vctr.z);
}

inline CVector& CVector::operator+=(CVector const& vctr){
	// iKłwlĂȂ
	x+=vctr.x;
	y+=vctr.y;
	z+=vctr.z;
	return *this;
}

inline CVector CVector::operator-(CVector const& vctr) const //fixed by kamiyan
{
	// iKłwlĂȂ
	return CVector(
		x - vctr.x,
		y - vctr.y,
		z - vctr.z);
}

inline CVector& CVector::operator-=(CVector const& vctr){
	// iKłwlĂȂ
	x-=vctr.x;
	y-=vctr.y;
	z-=vctr.z;
	return *this;
}

inline CVector CVector::operator*(double val) const
{
	// iKłwlĂȂ
	return CVector(
		x * val,
		y * val,
		z * val);
}

inline CVector& CVector::operator*=(double val)
{
	// iKłwlĂȂ
	x*=val;
	y*=val;
	z*=val;
	return *this;
}

inline CVector CVector::operator/(double val) const
{
	// iKłwlĂȂ
	return CVector(
		x / val,
		y / val,
		z / val);
}

inline CVector& CVector::operator/=(double val)
{
	// iKłwlĂȂ
	x/=val;
	y/=val;
	z/=val;
	return *this;
}

inline BOOL CVector::operator==(CVector const& vctr) const
{
	for(int row=0; row<4; row++){
		if(m_rElement[row]!=vctr.m_rElement[row]){
			return FALSE;
		}
	}
	return TRUE;
}

inline BOOL CVector::operator!=(CVector const& vctr) const
{
	return !((*this)==vctr);
}

inline double CVector::DotProduct(CVector const& vctr) const
{
	// iKłwlĂȂ
	return x*vctr.x +y*vctr.y + z*vctr.z;
}

inline CVector CVector::CrossProduct(CVector const& vctr) const
{
	// iKłwlĂȂ
	return CVector(
		y * vctr.z - z * vctr.y,
		z * vctr.x - x * vctr.z,
		x * vctr.y - y * vctr.x);
}

inline CVector operator*(double val, CVector const& vctr)
{
	// iKłwlĂȂ
	return CVector(
		val * vctr.x,
		val * vctr.y,
		val * vctr.z);
}

//CMatrixNX -----------------------------------------------------
class CMatrix {
public:
	double m_rElement[4][4];

public:
	CMatrix();
	CMatrix(CMatrix const& mtrx);
	~CMatrix(){}

	double GetElement(int row, int col) const;
	CMatrix& SetElement(int row, int col, double val);
	double& operator()(int row, int col);

	BOOL operator==(CMatrix const& mtrx) const;
	BOOL operator!=(CMatrix const& mtrx) const;
	CMatrix& operator=(CMatrix const& mtrx);
	CMatrix operator*(CMatrix const& mtrx) const;
	CMatrix& operator*=(CMatrix const& mtrx){
		return *this=(*this)*mtrx;
	}
	CMatrix& SetUnit();
	BOOL IsUnit() const;

	CMatrix& AddXRotation(double rot);
	CMatrix& SetXRotation(double rot);
	CMatrix& AddYRotation(double rot);
	CMatrix& SetYRotation(double rot);
	CMatrix& AddZRotation(double rot);
	CMatrix& SetZRotation(double rot);
	CMatrix& SetRotation(double x, double y, double z, double rot);
	CMatrix& SetRotation(CVector const& v,double rot){
		return SetRotation(v.x,v.y,v.z,rot);
	}
	CMatrix& AddRotation(double x, double y, double z, double rot){
		CMatrix m;
		m.SetRotation(x,y,z,rot);
		*this*=m;
		return *this;
	}
	CMatrix& AddRotation(CVector const& v,double rot){
		return AddRotation(v.x,v.y,v.z,rot);
	}
	CMatrix& AddTranslation(double x, double y, double z);
	CMatrix& AddTranslation(CVector const& v){
		return AddTranslation(v.x,v.y,v.z);
	}
	CMatrix& SetTranslation(double x, double y, double z);
	CMatrix& SetTranslation(CVector const& v){
		return SetTranslation(v.x,v.y,v.z);
	}
	CMatrix& AddScale(double x, double y, double z);
	CMatrix& AddScale(CVector const& v){
		return AddScale(v.x,v.y,v.z);
	}
	CMatrix& AddScale(double val){
		return AddScale(val,val,val);
	}
	CMatrix& SetScale(double x, double y, double z);
	CMatrix& SetScale(CVector const& v){
		return SetScale(v.x,v.y,v.z);
	}
	CMatrix& SetScale(double val){
		return SetScale(val,val,val);
	}
	CMatrix& InverseMatrix();

	friend CVector CVector::operator*(CMatrix const& mtrx) const;
};

inline CVector CVector::operator*(CMatrix const& mtrx) const
{
	CVector	vctr;
	for(int i=0; i<4; i++){
		double	val = double(0.0);		
		for(int j=0; j<4; j++){
			val += m_rElement[j] * mtrx.m_rElement[i][j];
		}
		vctr.m_rElement[i] = val;
	}
	return vctr;
}

inline CVector& CVector::operator*=(CMatrix const& mtrx)
{
	return (*this = (*this) * mtrx);
}


#endif //__VECTOR_H__
